/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// java dependencies
import java.util.HashMap;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * TODO:
 * docs
 */
public class Grid2DSquareCellDoubleIterator extends AbstractIterator {
    
    private Iterator grid2DSquareCellDoubleHashMapIterator;
    private Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk;
    private AbstractIterator grid2DSquareCellDoubleChunkIterator;
    
    /** Creates a new instance of Grid2DSquareDoubleIterator */
    public Grid2DSquareCellDoubleIterator() {}
    
    /**
     * Creates a new instance of Grid2DSquareDoubleIterator
     * @param grid2DSquareCellDouble The Grid2DSquareCellDouble to iterate over.
     */
    public Grid2DSquareCellDoubleIterator( Grid2DSquareCellDouble grid2DSquareCellDouble ) {
        this.grid2DSquareCellDoubleHashMapIterator = grid2DSquareCellDouble.grid2DSquareCellChunks.values().iterator();
        if ( grid2DSquareCellDoubleHashMapIterator.hasNext() ) {
            //this.grid2DSquareCellDoubleChunkIterator = ( ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next() );
            this.grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next();
            initGrid2DSquareCellDoubleChunkIterator();
        }
    }
    
    /**
     * Initialises this.grid2DSquareCellDoubleChunkIterator from this.grid2DSquareCellDoubleChunk
     */
    private void initGrid2DSquareCellDoubleChunkIterator() {
        try {
            this.grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunk64CellMapIterator( ( Grid2DSquareCellDoubleChunk64CellMap ) this.grid2DSquareCellDoubleChunk );
        } catch ( ClassCastException cce0 ) {
            try {
                this.grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkArrayIterator( ( Grid2DSquareCellDoubleChunkArray ) this.grid2DSquareCellDoubleChunk );
            } catch ( ClassCastException cce1 ) {
                try {
                    this.grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkJAIIterator( ( Grid2DSquareCellDoubleChunkJAI ) this.grid2DSquareCellDoubleChunk );
                } catch ( ClassCastException cce2 ) {
                    this.grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkMapIterator( ( Grid2DSquareCellDoubleChunkMap ) this.grid2DSquareCellDoubleChunk );
                }
            }
        }
    }
    
    /**
     * Initialises this.grid2DSquareCellDoubleChunkIterator from this.grid2DSquareCellDoubleChunk
     */
    public AbstractIterator initGrid2DSquareCellDoubleChunkIterator( Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk ) {
        AbstractIterator grid2DSquareCellDoubleChunkIterator;
        try {
            grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunk64CellMapIterator( ( Grid2DSquareCellDoubleChunk64CellMap ) grid2DSquareCellDoubleChunk );
        } catch ( ClassCastException cce0 ) {
            try {
                grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkArrayIterator( ( Grid2DSquareCellDoubleChunkArray ) grid2DSquareCellDoubleChunk );
            } catch ( ClassCastException cce1 ) {
                try {
                    grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkJAIIterator( ( Grid2DSquareCellDoubleChunkJAI ) grid2DSquareCellDoubleChunk );
                } catch ( ClassCastException cce2 ) {
                    grid2DSquareCellDoubleChunkIterator = new Grid2DSquareCellDoubleChunkMapIterator( ( Grid2DSquareCellDoubleChunkMap ) grid2DSquareCellDoubleChunk );
                }
            }
        }
        return grid2DSquareCellDoubleChunkIterator;
    }
    
    /**
     * Returns <tt>true</tt> if the iteration has more elements. (In other
     * words, returns <tt>true</tt> if <tt>next</tt> would return an element
     * rather than throwing an exception.)
     *
     * @return <tt>true</tt> if the iterator has more elements.
     */
    public boolean hasNext() {
        if ( this.grid2DSquareCellDoubleChunkIterator.hasNext() ) {
            return true;
        } else {
            while ( grid2DSquareCellDoubleHashMapIterator.hasNext() ) {
                //this.grid2DSquareCellDoubleChunkIterator = ( ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next() );
                Grid2DSquareCellDoubleChunkAbstract grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next();
                AbstractIterator grid2DSquareCellDoubleChunkIterator = initGrid2DSquareCellDoubleChunkIterator( grid2DSquareCellDoubleChunk );
                if ( grid2DSquareCellDoubleChunkIterator.hasNext() ) {
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Returns the next element in the iteration.
     *
     * @return the next element in the iteration.
     * @exception NoSuchElementException iteration has no more elements.
     */
    public Object next() {
        if ( this.grid2DSquareCellDoubleChunkIterator.hasNext() ) {
            return this.grid2DSquareCellDoubleChunkIterator.next();
        } else {
            while ( grid2DSquareCellDoubleHashMapIterator.hasNext() ) {
                //this.grid2DSquareCellDoubleChunkIterator = ( ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next() );
                this.grid2DSquareCellDoubleChunk = ( Grid2DSquareCellDoubleChunkAbstract ) this.grid2DSquareCellDoubleHashMapIterator.next();
                this.grid2DSquareCellDoubleChunkIterator = initGrid2DSquareCellDoubleChunkIterator( grid2DSquareCellDoubleChunk );
                if ( this.grid2DSquareCellDoubleChunkIterator.hasNext() ) {
                    return this.grid2DSquareCellDoubleChunkIterator.next();
                }
            }
        }
        // Should this be an error? or null or what
        return new NoSuchElementException();
        //return null;
    }
    
    /**
     *
     * Removes from the underlying collection the last element returned by the
     * iterator (optional operation).  This method can be called only once per
     * call to <tt>next</tt>.  The behavior of an iterator is unspecified if
     * the underlying collection is modified while the iteration is in
     * progress in any way other than by calling this method.
     *
     * @exception UnsupportedOperationException if the <tt>remove</tt>
     *		  operation is not supported by this Iterator.
     *
     * @exception IllegalStateException if the <tt>next</tt> method has not
     *		  yet been called, or the <tt>remove</tt> method has already
     *		  been called after the last call to the <tt>next</tt>
     *		  method.
     */
    public void remove() {}
    
}
