/**
 * Version 1.0 is to handle individual single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.core;
// gnu.trove dependencies
//import gnu.trove.TDoubleIntHashMap;
//import gnu.trove.TDoubleIntIterator;
//import gnu.trove.TDoubleIntProcedure;
//import gnu.trove.TDoubleLongHashMap;
//import gnu.trove.TIntProcedure;
// java dependencies
import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Arrays;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract.ChunkID;

/**
 * TODO:
 * docs
 */
public class Grid2DSquareCellIntChunkArray extends Grid2DSquareCellIntChunkAbstract implements Serializable {
    
    private static final long serialVersionUID = 1L; 
    
    /**
     * For storing values arranged in rows and columns.
     */
    private int[][] data;
    
    /**
     * Default constructor
     */
    protected Grid2DSquareCellIntChunkArray() {}
 
    /**
     * Creates a new Grid2DSquareCellInt grid containing all no data values.
     * @param grid2DSquareCellInt
     * @param chunkID
     */
    protected Grid2DSquareCellIntChunkArray( Grid2DSquareCellInt grid2DSquareCellInt, ChunkID chunkID ) {
        this.chunkID = chunkID;
        initGrid2DSquareCell( grid2DSquareCellInt );
        int noDataValue = grid2DSquareCellInt.getNoDataValue( grid2DSquareCellInt.handleOutOfMemoryErrorFalse );
        int chunkNrows = grid2DSquareCellInt.getChunkNrows( chunkID, grid2DSquareCellInt.handleOutOfMemoryErrorFalse );
        this.data = new int[ chunkNrows ][ grid2DSquareCellInt.getChunkNcols( chunkID, grid2DSquareCellInt.handleOutOfMemoryErrorFalse ) ];
        int row;
        for ( row = 0; row < chunkNrows; row ++ ) {
            Arrays.fill( data[ row ], noDataValue );
        }
    }
    
    /**
     * TODO:
     * 1. docs
     * 2. A fast toArray() method in Grid2DSquareCellIntChunkMap could be coded 
     * then a constructor based on an int[] or int[][] might be faster?
     */
    protected Grid2DSquareCellIntChunkArray( Grid2DSquareCellIntChunkAbstract grid2DSquareCellIntChunk, ChunkID chunkID ) {
        this.chunkID = chunkID;
        Grid2DSquareCellInt grid2DSquareCellInt = grid2DSquareCellIntChunk.getGrid2DSquareCellInt();
        initGrid2DSquareCell( grid2DSquareCellInt );
        int chunkNrows = grid2DSquareCellInt.getChunkNrows( chunkID, grid2DSquareCellInt.handleOutOfMemoryErrorFalse );
        int chunkNcols = grid2DSquareCellInt.getChunkNcols( chunkID, grid2DSquareCellInt.handleOutOfMemoryErrorFalse );
        int noDataValue = grid2DSquareCellInt.getNoDataValue( grid2DSquareCellInt.handleOutOfMemoryErrorFalse );
        initData();
        int row;
        int col;
        for ( row = 0; row < chunkNrows; row ++ ) {
            for ( col = 0; col < chunkNcols; col ++ ) {
                this.data[ row ][ col ] = grid2DSquareCellIntChunk.getCell( row, col, noDataValue );
                //initCell( row, col, grid2DSquareCellIntChunk.getCell( row, col ) );
            }
        }
    }
    
    /**
     * Returns a description of this.
     */
    public String toString() {
        return "Grid2DSquareCellIntChunkArray( ChunkID( " + this.chunkID.toString() + " ) )";
    }
    
    /**
     * Initialises the data associated with this.
     */
    public void initData() {
        Grid2DSquareCellInt grid2DSquareCellInt = this.getGrid2DSquareCellInt();
        int chunkNcols = grid2DSquareCellInt.getChunkNcols(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        int chunkNrows = grid2DSquareCellInt.getChunkNrows(grid2DSquareCellInt.handleOutOfMemoryErrorFalse);
        this.data = new int[ chunkNrows ][ chunkNcols ];
    }
    
    /**
     * TODO: docs
     */
    public int[][] getData(){
        return this.data;
    }
    
    /**
     * Clears the data associated with this.
     */
    public void clearData() {
        this.data = null;
        System.gc();
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell row chunkCellColIndex.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param noDataValue the noDataValue of this.grid2DSquareCellInt
     */
    public int getCell( int chunkCellRowIndex, int chunkCellColIndex, int noDataValue ) {
        try {
            return this.data[ chunkCellRowIndex ][ chunkCellColIndex ];
        } catch ( Exception e0 ) {
            return noDataValue;
        }
    }
    
    /**
     * Initialises the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell column chunkCellColIndex. Utility method for constructor.
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToInitialise the value with which the cell is initialised
     */
    protected void initCell(int chunkCellRowIndex, int chunkCellColIndex, int valueToInitialise) {
        this.data[ chunkCellRowIndex ][ chunkCellColIndex ] = valueToInitialise;
    }
    
    /**
     * Returns the value at position given by: chunk cell row chunkCellRowIndex; 
     * chunk cell row chunkCellColIndex and sets it to valueToSet
     * @param chunkCellRowIndex the row index of the cell w.r.t. the origin of this chunk
     * @param chunkCellColIndex the column index of the cell w.r.t. the origin of this chunk
     * @param valueToSet the value the cell is to be set to
     * @param noDataValue the noDataValue of this.grid2DSquareCellInt
     */
    public int setCell( int chunkCellRowIndex, int chunkCellColIndex, int valueToSet, int noDataValue ) {
        try {
            int oldValue = this.data[ chunkCellRowIndex ][ chunkCellColIndex ];
            this.data[ chunkCellRowIndex ][ chunkCellColIndex ] = valueToSet;
            return oldValue;
        } catch ( Exception e0 ) {
            return noDataValue;
        }
    }
    
    /**
     * Returns a Grid2DSquareCellIntChunkArrayIterator for iterating over 
     * the cells in this.
     */
    public AbstractIterator iterator() {
        return new Grid2DSquareCellIntChunkArrayIterator( this );
    }
    
}