/**
 * Version 1.0 is utilities with general utility external to uk.ac.leeds.ccg.andyt.grids.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.utilities;
// java dependencies
import java.io.Serializable;
import java.util.Arrays;

/**
 * A class to be extended by classes that handle runtime errors such as 
 * java.lang.OutOfMemroyError
 * TODO:
 * Add a constant int TYPE which can be used to switch between different types 
 * of this.
 * Handling for NumberFormatExceptions and ArithmeticExceptions in calculations
 */
public abstract class ErrorHandler implements Serializable {
    
    private static final long serialVersionUID = 1L;
    
    public static final boolean handleOutOfMemoryErrorFalse = false;

    public static final boolean handleOutOfMemoryErrorTrue = true;
    
    /**
     * Reserve memory to be cleared when dealing with java.lang.OutOfMemoryErrors.
     */
    protected int[] memoryReserve;
    //protected Integer[] memoryReserve;
    
    /**
     * Initialises memoryReserve as an array of size.
     * @param size The size that memoryReserve is initialised to.
     */
    public final void initMemoryReserve( int size ) {
        //log( "initMemoryReserve( " + size + " )" );
        this.memoryReserve = new int[ size ];
        Arrays.fill( this.memoryReserve, Integer.MIN_VALUE );
        System.gc();
    }
    
    /**
     * Initialises memoryReserve as an array of size 2000000.
     * More than 1000000 is needed for caching and clearing chunks in 
     * Grid2DSquareCellAbstract constructors.
     */
    public final void initMemoryReserve() {
        initMemoryReserve( 5000000 );
    }
    
    /**
     * Clears memoryReserve by setting it to null and calling the garbage
     * collector.
     */
    public final void clearMemoryReserve() {
        //System.gc();
        //log(Runtime.getRuntime().freeMemory());
        this.memoryReserve = null;
        System.gc();
        //log(Runtime.getRuntime().freeMemory());
    }
    
}