/**
 * Version 1.0 is utilities with general utility external to uk.ac.leeds.ccg.andyt.grids.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.utilities;
// java dependencies
import java.awt.geom.Point2D;
// uk.ac.leeds.ccg.andyt.grids dependencies
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;

/**
 * Class of methods to do with kernels.
 */
public abstract class Kernel {
    
    /**
     * Returns a double value for the height of a kernel at thisDistance from
     * the centre of a kernel with; Bandwidth distance, weight at the centre of
     * weightIntersect and distance decay of weightFactor.
     * TODO: If weightfactor is < 1.0 strange things could be happening!!!!
     */
    public static double getKernelWeight( double distance, double weightIntersect, double weightFactor, double thisDistance ) {
        double weight;
        // The following weight is just one example of a kernel that can be used!
        // It provides a general monotonic curve based on distance over bandwidth.
        /*
        if ( weightFactor > 0.0d ) {
            weight = ( Math.pow( 1.0d - ( Math.pow( thisDistance , 2.0d ) / Math.pow( distance, 2.0d ) ), weightFactor ) * weightIntersect );
            //weight = Math.pow( 1.0d - ( Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor ) ), distanceWeightFactor );
        } else {
            if ( weightFactor < 0.0d ) {
                weight = ( ( 1.0d - Math.pow( 1.0d - ( Math.pow( thisDistance , 2.0d ) / Math.pow( distance, 2.0d ) ), Math.abs( weightFactor ) ) * ( 1.0d - weightIntersect ) ) + weightIntersect );
                //weight = 1.0d - Math.pow( 1.0d - ( Math.pow( thisDistance, Math.abs( distanceWeightFactor ) ) / Math.pow( distance, Math.abs( distanceWeightFactor ) ) ), Math.abs( distanceWeightFactor ) );
            } else {
                weight = weightIntersect;
            }
        }
        return weight;
         */
        //if ( distance == thisDistance ) { return 0.0d; }
        return Math.pow( 1.0d - ( Math.pow( thisDistance , 2.0d ) / Math.pow( distance, 2.0d ) ), weightFactor ) * weightIntersect;
        //return Math.pow( 1.0d - ( Math.pow( thisDistance, distanceWeightFactor ) / Math.pow( distance, distanceWeightFactor ) ), distanceWeightFactor );
    }
    
    /**
     * Returns a double[][] of values for a kernel
     */
    public static double[][] getKernelWeights( Grid2DSquareCellDouble grid, double distance, double weightIntersect, double weightFactor ) {
        double cellsize = grid.getDimensions()[0].doubleValue();
        double noDataValue = grid.getNoDataValue();
        int cellDistance = ( int ) Math.ceil( distance / cellsize );
        double[][] weights = new double[ ( cellDistance * 2 ) + 1 ][ ( cellDistance * 2 ) + 1 ];
        // The following weight is just one example of a kernel that can be used!
        // It provides a general monotonic curve based on distance over bandwidth.
        boolean handleOutOfMemroyError = true;
        double x0 = grid.getCellX( 0L, handleOutOfMemroyError );
        double y0 = grid.getCellY( 0L, handleOutOfMemroyError );
        double x1;
        double y1;
        double thisDistance;
        int row;
        int col;
        for ( row = - cellDistance; row <= cellDistance; row ++ ) {
            for ( col = - cellDistance; col <= cellDistance; col ++ ) {
                x1 = grid.getCellX( col, handleOutOfMemroyError );
                y1 = grid.getCellY( row, handleOutOfMemroyError );
                thisDistance = Utilities.distance( x0, y0, x1, y1 );
                //if ( thisDistance <= distance ) {
                if ( thisDistance < distance ) {
                    weights[ row + cellDistance ][ col + cellDistance ] = getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
                } else {
                    //weights[ i + cellDistance ][ j + cellDistance ] = noDataValue;
                    weights[ row + cellDistance ][ col + cellDistance ] = 0.0d;
                }
            }
        }
        return weights;
    }
    
    /**
     * Returns a double[] of values for a kernel
     * TODO: document
     * @param grid
     */
    public static double[] getKernelWeights( Grid2DSquareCellDouble grid, long rowIndex, long colIndex, double distance, double weightIntersect, double weightFactor, Point2D.Double[] points ) {
        double[] weights = new double[ points.length ];
        // The following weight is just one example of a kernel that can be used!
        // It provides a general monotonic curve based on distance over bandwidth.
        boolean handleOutOfMemroyError = true;
        Point2D.Double centroid = new Point2D.Double( grid.getCellX( colIndex, handleOutOfMemroyError ), grid.getCellY( rowIndex, handleOutOfMemroyError ) );
        double thisDistance;
        double weight;
        for ( int i = 0; i < points.length; i ++ ) {
            thisDistance = centroid.distance( points[ i ] );
            if ( thisDistance < distance ) {
                weights[ i ] = getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
            }
        }
        return weights;
    }
    
    /**
     * TODO: document
     */
    public static double[] getKernelWeights( Point2D.Double centroid, double distance, double weightIntersect, double weightFactor, Point2D.Double[] points ) {
        double[] weights = new double[ points.length ];
        // The following weight is just one example of a kernel that can be used!
        // It provides a general monotonic curve based on distance over bandwidth.
        double thisDistance;
        for ( int i = 0; i < points.length; i ++ ) {
            thisDistance = centroid.distance( points[ i ] );
            if ( thisDistance < distance ) {
                weights[ i ] = getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
            }
        }
        return weights;
    }
    
    
    /**
     * Returns double[] result:
     * double[0] the total sum of all the weights for a given kernel
     * double[1] the total number of cells thats centroids are within distance
     * of an arbitrary cell centroid of grid0
     * @param grid0
     * @param grid0CellDistance
     * @param distance
     * @param weightIntersect
     * @param weightFactor
     */
    public static double[] getKernelParameters( Grid2DSquareCellDouble grid0, int grid0CellDistance, double distance, double weightIntersect, double weightFactor ) {
        double kernelParameters[] = new double[ 2 ];
        kernelParameters[ 0 ] = 0.0d;
        kernelParameters[ 1 ] = 0.0d;
        boolean handleOutOfMemroyError = true;
        double x0 = grid0.getCellX( 0, handleOutOfMemroyError );
        double y0 = grid0.getCellY( 0, handleOutOfMemroyError );
        double x1;
        double y1;
        double thisDistance;
        for ( int p = - grid0CellDistance; p <= grid0CellDistance; p ++ ) {
            for ( int q = - grid0CellDistance; q <= grid0CellDistance; q ++ ) {
                x1 = grid0.getCellX( q, handleOutOfMemroyError );
                y1 = grid0.getCellY( p, handleOutOfMemroyError );
                thisDistance = Utilities.distance( x0, y0, x1, y1 );
                //if ( thisDistance <= distance ) {
                if ( thisDistance < distance ) {
                    kernelParameters[ 0 ] += getKernelWeight( distance, weightIntersect, weightFactor, thisDistance );
                    kernelParameters[ 1 ] += 1.0d;
                }
            }
        }
        return kernelParameters;
    }
    
    public static double getAdaptiveKernelWeight( double distance, double bandwidth, double sumWeights, int precision, double weightIntersect, double weightFactor ) {
        //int sectionSize = ( int ) Math.ceil( bandwidth / ( double ) precision );
        //double sectionArea = Math.pow( Math.ceil( bandwidth / ( double ) precision ), 2.0d );
        double kernelVolume = getKernelVolume( bandwidth, precision, weightIntersect, weightFactor );
        double kernelWeight = getKernelWeight( bandwidth, weightIntersect, weightFactor, distance );
        return ( kernelWeight * sumWeights ) / kernelVolume;
    }
    
    /**
     *
     */
    public static double getKernelVolume( double bandwidth, int precision, double weightIntersect, double weightFactor ) {
        double sumKernelWeights = 0.0d;
        int sectionSize = ( int ) Math.ceil( bandwidth / ( double ) precision );
        double sectionArea = ( double ) sectionSize * sectionSize;
        double thisDistance;
        Point2D.Double kernelCentroid = new Point2D.Double( 0.0d, 0.0d );
        Point2D.Double point;
        //int sectionCount = 0;
        for ( int p = 1; p < precision; p ++ ) {
            for ( int q = 0; q < precision; q ++ ) {
                point = new Point2D.Double( ( double ) p * sectionSize, ( double ) q * sectionSize );
                thisDistance = point.distance( kernelCentroid );
                if ( thisDistance < bandwidth ) {
                    sumKernelWeights += getKernelWeight( bandwidth, weightIntersect, weightFactor, thisDistance );
                    //sectionCount ++;
                }
            }
        }
        // Multiply by 4 for all quadrants
        // Add kernelCentroid weight
        // Multiply result be sectionArea to get volume
        return ( ( sumKernelWeights * 4.0d ) + weightIntersect ) * sectionArea;
    }
    
}
