/*
 * ESRIAsciigridImporterTest.java
 * JUnit based test
 */

package uk.ac.leeds.ccg.andyt.grids.exchange;

import junit.framework.*;
import java.io.File;
import java.io.StreamTokenizer;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.math.BigDecimal;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkAbstractFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleChunkRAFFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDoubleFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellInt;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunk64CellMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkAbstractFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkArrayFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkJAIFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkMapFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntChunkRAFFactory;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellIntFactory;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatistics0;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatistics1;
import uk.ac.leeds.ccg.andyt.grids.core.GridStatisticsAbstract;

public class ESRIAsciigridImporterTest extends TestCase {
    
    public ESRIAsciigridImporterTest(String testName) {
        super(testName);
    }
    
    ESRIAsciiGridImporter eSRIAsciigridImporter = new ESRIAsciiGridImporter();
    
    protected void setUp() throws java.lang.Exception {
        eSRIAsciigridImporter = new ESRIAsciiGridImporter( new File( "C:/Work/src/andyt/java/grids/data/TestData/ESRIAsciigrid/test0.asc" ) );
    }
    
    protected void tearDown() throws java.lang.Exception {
    }
    
    public static junit.framework.Test suite() {
        junit.framework.TestSuite suite = new junit.framework.TestSuite(ESRIAsciigridImporterTest.class);
        
        return suite;
    }
    
    /**
     * Test of readHeaderDouble method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testReadHeaderDouble() {
        System.out.println("testReadHeaderDouble");
        
        // TODO add your test code below by replacing the default call to fail.
        double[] header = eSRIAsciigridImporter.readHeaderDouble();
        //System.out.println( "ncols = " + header[ 0 ] );
        //System.out.println( "nrows = " + header[ 1 ] );
        //System.out.println( "xllcorner = " + header[ 2 ] );
        //System.out.println( "yllcorner = " + header[ 3 ] );
        //System.out.println( "cellsize = " + header[ 4 ] );
        //System.out.println( "NODATA_VALUE = " + header[ 5 ] );
        if ( ! ( header[ 0 ] == 160 &&
                header[ 1 ] == 230 &&
                header[ 2 ] == 389917.53125 &&
                header[ 3 ] == 1105108.129883 &&
                header[ 4 ] == 500.0 &&
                header[ 5 ] == -9999.0 ) ) {
            fail();
        }
    }
    
    /**
     * Test of readHeaderObject method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testReadHeaderObject() {
        System.out.println("testReadHeaderObject");
        Object[] header = eSRIAsciigridImporter.readHeaderObject();
        if ( ! ( header[ 0 ].getClass() == Long.class &&
                header[ 1 ].getClass() == Long.class &&
                header[ 2 ].getClass() == BigDecimal.class &&
                header[ 3 ].getClass() == BigDecimal.class &&
                header[ 4 ].getClass() == BigDecimal.class &&
                header[ 5 ].getClass() == Double.class ) ) {
            fail();
        }
    }
    
    /**
     * Test of readDouble method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testReadDouble() {
        double[] header = eSRIAsciigridImporter.readHeaderDouble();
        long numberOfCells = ( long ) header[ 0 ] * ( long ) header[ 1 ];
        //System.out.println( "numberOfCells = " + numberOfCells );
        double value;
        for ( long i = 0L; i < numberOfCells; i ++ ) {
            value = eSRIAsciigridImporter.readDouble();
            if ( i == 1257 ) {
                if ( value != header[ 5 ] ) {
                    fail();
                }
            } else {
                if ( i == 1258 ) {
                    if ( value != 0.0d ) {
                        fail();
                    }
                } else {
                    if ( i == 1266 ) {
                        if ( value != 1.0d ) {
                            fail();
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Test of readInt method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testReadInt() {
        System.out.println("testReadInt");
        double[] header = eSRIAsciigridImporter.readHeaderDouble();
        long numberOfCells = ( long ) header[ 0 ] * ( long ) header[ 1 ];
        //System.out.println( "numberOfCells = " + numberOfCells );
        int value;
        for ( long i = 0L; i < numberOfCells; i ++ ) {
            value = eSRIAsciigridImporter.readInt();
            if ( i == 1257 ) {
                if ( value != ( int ) header[ 5 ] ) {
                    fail();
                }
            } else {
                if ( i == 1258 ) {
                    if ( value != 0 ) {
                        fail();
                    }
                } else {
                    if ( i == 1266 ) {
                        if ( value != 1 ) {
                            fail();
                        }
                    }
                }
            }
        }
    }

    /**
     * Test of constructGrid2DSquareCellInt method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testConstructGrid2DSquareCellInt() {
        System.out.println("testConstructGrid2DSquareCellInt");
        
        // Initialisation
        int chunkNrows = 200; //2000
        int chunkNcols = 200; //2000
        // Chunk factories
        Grid2DSquareCellIntChunk64CellMapFactory grid2DSquareCellInt64CellMapArrayFactory = new Grid2DSquareCellIntChunk64CellMapFactory();
        Grid2DSquareCellIntChunkArrayFactory grid2DSquareCellIntChunkArrayFactory = new Grid2DSquareCellIntChunkArrayFactory();
        Grid2DSquareCellIntChunkJAIFactory grid2DSquareCellIntChunkJAIFactory = new Grid2DSquareCellIntChunkJAIFactory();
        Grid2DSquareCellIntChunkMapFactory grid2DSquareCellIntChunkMapFactory = new Grid2DSquareCellIntChunkMapFactory();
        Grid2DSquareCellIntChunkRAFFactory grid2DSquareCellIntChunkRAFFactory = new Grid2DSquareCellIntChunkRAFFactory();
        Grid2DSquareCellIntChunkAbstractFactory grid2DSquareCellIntChunkFactory = grid2DSquareCellIntChunkArrayFactory;
        // GridStatistics
        GridStatistics0 gridStatistics0 = new GridStatistics0();
        GridStatistics1 gridStatistics1 = new GridStatistics1();
        GridStatisticsAbstract gridStatistics = gridStatistics0;
        
        File outputDirectory = new File( "C:/Work/src/andyt/java/grids/data/TestData/Grid2DSquareCellInt/"+eSRIAsciigridImporter.getFilenamePrefix()+"_chunkNrows("+chunkNrows+")_chunkNcols("+chunkNcols+")" );
        
        Grid2DSquareCellInt grid2DSquareCellInt = this.eSRIAsciigridImporter.constructGrid2DSquareCellInt( gridStatistics, outputDirectory, grid2DSquareCellIntChunkFactory, chunkNrows, chunkNcols );
        
        System.out.println( grid2DSquareCellInt.toString() );
    }

    /**
     * Test of constructGrid2DSquareCellDouble method, of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testConstructGrid2DSquareCellDouble() {
        System.out.println("testConstructGrid2DSquareCellDouble");
        
        // Initialisation
        int chunkNrows = 200; //2000
        int chunkNcols = 200; //2000
        // Chunk factories
        Grid2DSquareCellDoubleChunk64CellMapFactory grid2DSquareCellDouble64CellMapArrayFactory = new Grid2DSquareCellDoubleChunk64CellMapFactory();
        Grid2DSquareCellDoubleChunkArrayFactory grid2DSquareCellDoubleChunkArrayFactory = new Grid2DSquareCellDoubleChunkArrayFactory();
        Grid2DSquareCellDoubleChunkJAIFactory grid2DSquareCellDoubleChunkJAIFactory = new Grid2DSquareCellDoubleChunkJAIFactory();
        Grid2DSquareCellDoubleChunkMapFactory grid2DSquareCellDoubleChunkMapFactory = new Grid2DSquareCellDoubleChunkMapFactory();
        Grid2DSquareCellDoubleChunkRAFFactory grid2DSquareCellDoubleChunkRAFFactory = new Grid2DSquareCellDoubleChunkRAFFactory();
        Grid2DSquareCellDoubleChunkAbstractFactory grid2DSquareCellDoubleChunkFactory = grid2DSquareCellDoubleChunkArrayFactory;
        // GridStatistics
        GridStatistics0 gridStatistics0 = new GridStatistics0();
        GridStatistics1 gridStatistics1 = new GridStatistics1();
        GridStatisticsAbstract gridStatistics = gridStatistics0;
        
        File outputDirectory = new File( "C:/Work/src/andyt/java/grids/data/TestData/Grid2DSquareCellDouble/"+eSRIAsciigridImporter.getFilenamePrefix()+"_chunkNrows("+chunkNrows+")_chunkNcols("+chunkNcols+")" );
        
        Grid2DSquareCellDouble grid2DSquareCellDouble = this.eSRIAsciigridImporter.constructGrid2DSquareCellDouble( gridStatistics, outputDirectory, grid2DSquareCellDoubleChunkFactory, chunkNrows, chunkNcols );
        
        System.out.println( grid2DSquareCellDouble.toString() );
    }

    /**
     * Untested methods of class uk.ac.leeds.ccg.andyt.grids.exchange.ESRIAsciigridImporter.
     */
    public void testClose() {
        System.out.println("testClose");        
    }

}
